import inspect
from typing import Callable, Dict, Any

from dataclasses import dataclass


from petisco.events.publisher.domain.interface_event_publisher import IEventPublisher
from petisco.events.service_deployed import ServiceDeployed
from petisco.events.subscriber.domain.interface_event_subscriber import IEventSubscriber
from petisco.frameworks.interface_application import IApplication
from petisco.logger.interface_logger import INFO, ILogger
from petisco.logger.log_message import LogMessage
from petisco.logger.not_implemented_logger import NotImplementedLogger
from petisco.application.config.config import Config
from petisco.application.singleton import Singleton
from petisco.application.interface_repository import IRepository
from petisco.application.interface_service import IService
from petisco.tasks.infrastructure.apscheduler_task_executor import (
    APSchedulerTaskExecutor,
)
from petisco import __version__


@dataclass
class Petisco(metaclass=Singleton):
    app_name: str
    app_version: str
    logger: ILogger = NotImplementedLogger()
    application: IApplication = None
    services_provider: Callable = None
    repositories_provider: Callable = None
    options: Dict[str, Any] = None
    info: Dict = None
    _persistence_models: Dict[str, Any] = False
    persistence_configured: bool = False
    config: Config = None
    event_publisher: IEventPublisher = None
    event_subscriber: IEventSubscriber = None

    def __init__(self, config: Config):
        self.config = config
        self.app_name = config.app_name
        self.app_version = config.app_version
        self.logger = config.get_logger()
        self.info = {
            "app_name": self.app_name,
            "app_version": self.app_version,
            "petisco_version": __version__,
        }
        self._set_persistence()
        self._set_providers()
        self._set_events()
        self.set_tasks()
        self.options = config.options

    @staticmethod
    def get_instance():
        try:
            return Petisco()
        except Exception as e:  # noqa E722
            frame_info = inspect.stack()[1]
            raise ImportError(
                f"Petisco must be configured when the application starts. \n"
                f"Petisco has been requested before its initial configuration.\n "
                f"Following code must be executed after Petisco initialization:\n"
                f"\tfilename: {frame_info.filename}\n"
                f"\tlineno: {frame_info.lineno}\n"
                f"\tfunction: {frame_info.function}\n"
                f"\tcode_context: {frame_info.code_context}"
            )

    @staticmethod
    def from_filename(filename: str):
        """
       Parameters
       ----------
       filename
           YAML-based configuration file (default petisco.yml)
        """
        print(f"Loading petisco from: {filename}")

        config = Config.from_filename(filename).unwrap_or_throw()

        return Petisco(config=config)

    def publish_deploy_event(self):
        if self.config.config_events.publish_deploy_event:
            event = ServiceDeployed(
                app_name=self.app_name, app_version=self.app_version
            )
            self.event_publisher.publish(event)

    def set_tasks(self):
        config_tasks = self.config.config_tasks
        if config_tasks.tasks:
            self.info["tasks"] = {}
            for task_name, config_task in config_tasks.tasks.items():
                self.info["tasks"][task_name] = config_task.to_dict()
            self.task_executor = APSchedulerTaskExecutor()

    def _schedule_tasks(self):
        config_tasks = self.config.config_tasks
        if config_tasks.tasks:
            self.task_executor.start(config_tasks)

    def _unschedule_tasks(self):
        self.task_executor.stop()

    def _set_persistence(self):
        self._persistence_models = {}
        config_persistence = self.config.config_persistence
        if config_persistence.config:
            import_database_models_func = (
                config_persistence.get_import_database_models_func()
            )
            config_persistence.config(import_database_models_func)
            self.persistence_configured = True
            self._persistence_models = config_persistence.get_models()

    def _set_providers(self):
        config_providers = self.config.config_providers
        if not config_providers:
            return
        if config_providers.config_dependencies:
            config_providers.config_dependencies()

        if config_providers.services_provider:
            self.services_provider = config_providers.services_provider

            info_services = {}
            for key, service in self.services_provider().items():
                if hasattr(service, "info"):
                    info_services[key] = service.info()
                else:
                    raise TypeError(
                        f"Service with key {key} ({type(service)}) must implement info"
                    )
            self.info["services"] = info_services
        if config_providers.repositories_provider:
            self.repositories_provider = config_providers.repositories_provider

            info_repositories = {}
            for key, repository in self.repositories_provider().items():
                if hasattr(repository, "info"):
                    info_repositories[key] = repository.info()
                else:
                    raise TypeError(
                        f"Repository with key {key} ({type(repository)}) must implement info"
                    )

            self.info["repositories"] = info_repositories

    def _set_events(self):
        config_events = self.config.config_events
        if not config_events:
            return

        if config_events.config_event_publisher.provider:
            self.event_publisher = config_events.config_event_publisher.provider()
            if hasattr(self.event_publisher, "info"):
                self.info["event_publisher"] = self.event_publisher.info()
            else:
                raise TypeError(
                    f"Given event_publisher ({type(self.event_publisher)}) must implement info"
                )

        if config_events.config_event_subscriber.provider:
            subscribers = config_events.config_event_subscriber.subscribers
            self.event_subscriber = config_events.config_event_subscriber.provider(
                subscribers
            )
            if hasattr(self.event_subscriber, "info"):
                self.info["event_subscriber"] = self.event_subscriber.info()
            else:
                raise TypeError(
                    f"Given event_subscriber ({type(self.event_subscriber)}) must implement info"
                )

    def _log_status(self):
        if self.info:
            self.logger.log(INFO, LogMessage(data={"message": {"info": self.info}}))

        if self.options:
            self.logger.log(
                INFO, LogMessage(data={"message": {"options": self.options}})
            )

    def _start(self):
        self.event_subscriber.start()
        self._schedule_tasks()
        self._log_status()
        self.publish_deploy_event()

    def start(self):
        self._start()
        self.config.get_application().start()

    def get_app(self):
        self._start()
        return self.config.get_application().get_app()

    def stop(self):
        self.event_subscriber.stop()
        self._unschedule_tasks()

    @staticmethod
    def services() -> Dict[str, IService]:
        return Petisco.get_instance().services_provider()

    @staticmethod
    def repositories() -> Dict[str, IRepository]:
        return Petisco.get_instance().repositories_provider()

    @staticmethod
    def get_service(key: str) -> IService:
        service = Petisco.services().get(key)
        if not service:
            raise ValueError(
                f"Petisco: {key} service is not defined. Please, add it to petisco.yml"
            )
        return service

    @staticmethod
    def get_repository(key: str) -> IRepository:
        repository = Petisco.repositories().get(key)
        if not repository:
            raise ValueError(
                f"Petisco: {key} repository is not defined. Please, add it to petisco.yml"
            )
        return repository

    @staticmethod
    def persistence_models() -> Dict[str, str]:
        persistence_models = {}
        try:
            persistence_models = Petisco.get_instance()._persistence_models
        except:  # noqa E722
            pass
        return persistence_models

    @staticmethod
    def get_persistence_model(key: str) -> Any:
        return Petisco.persistence_models().get(key)

    @staticmethod
    def persistence_session_scope():
        from petisco.persistence.sqlalchemy.sqlalchemy_session_scope import (
            session_scope,
        )

        return session_scope

    @staticmethod
    def providers():
        return Petisco.services(), Petisco.repositories()

    @staticmethod
    def get_event_publisher():
        return Petisco.get_instance().event_publisher

    @staticmethod
    def get_event_subscriber():
        return Petisco.get_instance().event_subscriber

    @staticmethod
    def get_logger():
        return Petisco.get_instance().logger

    @staticmethod
    def get_app_name():
        return Petisco.get_instance().app_name

    @staticmethod
    def get_app_version():
        return Petisco.get_instance().app_version



from typing import Dict

from petisco.events.event import Event
from petisco.events.publisher.domain.interface_event_publisher import IEventPublisher


class NotImplementedEventPublisher(IEventPublisher):
    def info(self) -> Dict:
        return {"name": self.__class__.__name__}

    def publish(self, event: Event):
        pass



import json
from typing import Dict

from healthcheck import EnvironmentDump
from meiga import Result, Error, Success, Failure

from petisco.application.petisco import Petisco
from petisco.modules.environment.domain.environment_provider_error import (
    EnvironmentProviderError,
)
from petisco.use_case.use_case import UseCase


class EnvironmentProvider(UseCase):
    def __init__(
        self,
        include_python: bool = True,
        include_os: bool = True,
        include_process: bool = True,
    ):
        self.include_python = include_python
        self.include_os = include_os
        self.include_process = include_process

    def execute(self, petisco: Petisco) -> Result[Dict, Error]:
        try:
            application_info = {"config": petisco.info}
            envdump = EnvironmentDump(
                include_python=self.include_python,
                include_os=self.include_os,
                include_process=self.include_process,
            )
            envdump.add_section("Environment", application_info)
            envdump_result = envdump.run()
        except Exception as e:
            return Failure(EnvironmentProviderError(str(e)))
        return Success(json.loads(envdump_result[0]))



import time


def timer(func):
    def _timer(*args, **kwargs):
        start_time = time.time()

        result = func(*args, **kwargs)

        elapsed_time = time.time() - start_time

        return result, elapsed_time

    return _timer



import logging
from typing import Callable

from petisco.logger.log_message import LogMessage
from petisco.logger.interface_logger import (
    ILogger,
    INFO,
    ERROR,
    WARN,
    WARNING,
    DEBUG,
    CRITICAL,
    FATAL,
)


class LoggingBasedLogger(ILogger):
    def __init__(
        self,
        logger_name: str,
        logging_level=logging.INFO,
        format: str = "%(name)s - %(levelname)s - %(message)s",
        config: Callable = None,
    ):

        logging.basicConfig(format=format, level=logging_level)
        if config:
            config()
        self.logger = logging.getLogger(logger_name)

    def log(self, logging_level, log_message: LogMessage):
        message = log_message.to_dict()
        if logging_level == INFO:
            message["meta"]["level"] = "info"
            self.logger.info(message)
        elif logging_level == DEBUG:
            message["meta"]["level"] = "debug"
            self.logger.debug(message)
        elif logging_level == CRITICAL:
            message["meta"]["level"] = "critical"
            self.logger.critical(message)
        elif logging_level == FATAL:
            message["meta"]["level"] = "fatal"
            self.logger.fatal(message)
        elif logging_level == ERROR:
            message["meta"]["level"] = "error"
            self.logger.error(message)
        elif logging_level == WARN:
            message["meta"]["level"] = "warning"
            self.logger.warn(message)
        elif logging_level == WARNING:
            message["meta"]["level"] = "warning"
            self.logger.warning(message)



import importlib
from typing import Optional, Dict, Any, Callable

from dataclasses import dataclass

from petisco.application.config.get_funtion_from_string import get_function_from_string
from petisco.application.config.raise_petisco_config_error import (
    raise_petisco_config_exception,
)


@dataclass
class ConfigPersistence:
    config: Optional[Callable] = None
    models: Optional[Dict[str, Any]] = None

    @staticmethod
    def from_dict(kdict):
        if not kdict or not isinstance(kdict, dict):
            return ConfigPersistence()
        config = (
            get_function_from_string(kdict.get("config"))
            .handle(
                on_failure=raise_petisco_config_exception,
                failure_args=(kdict, "persistence:config"),
            )
            .unwrap()
        )
        return ConfigPersistence(config=config, models=kdict.get("models"))

    def get_models(self):
        loaded_models = {}
        for name, model_string in self.models.items():
            mod_name, model_name = model_string.rsplit(".", 1)
            mod = importlib.import_module(mod_name)
            loaded_models[name] = getattr(mod, model_name)
        return loaded_models

    def get_import_database_models_func(self):
        def _import_database_models_func():
            for name, model_string in self.models.items():
                mod_name, model_name = model_string.rsplit(".", 1)
                __import__(mod_name, fromlist=[model_name])

        return _import_database_models_func



import os
import traceback
from typing import Dict, Any, List

import yaml
from meiga import Result, Error, Failure, Success
from meiga.decorators import meiga
from yaml.parser import ParserError, ScannerError

from petisco.application.config.config_persistence import ConfigPersistence
from petisco.application.config.config_providers import ConfigProviders
from petisco.tasks.config.config_tasks import ConfigTasks
from petisco.frameworks.interface_application import IApplication
from petisco.logger.interface_logger import ILogger
from petisco.logger.logging_based_logger import LoggingBasedLogger
from petisco.application.config.events.config_events import ConfigEvents
from petisco.application.config.config_file_not_found_error import (
    ConfigFileNotFoundError,
)
from petisco.application.config.config_file_not_valid_error import (
    ConfigFileNotValidError,
)
from petisco.application.config.config_framework import ConfigFramework
from petisco.application.config.config_logger import ConfigLogger
from petisco.logger.not_implemented_logger import NotImplementedLogger


class Config:
    def __init__(
        self,
        app_name: str,
        app_version: str,
        petisco_yml_folder: str = None,
        config_tasks: ConfigTasks = None,
        config_framework: ConfigFramework = None,
        config_logger: ConfigLogger = None,
        config_persistence: ConfigPersistence = None,
        config_providers: ConfigProviders = None,
        config_events: ConfigEvents = None,
        options: Dict = None,
    ):
        self.app_name = app_name
        self.app_version = app_version
        self.petisco_yml_folder = petisco_yml_folder
        self.config_tasks = config_tasks
        self.config_framework = config_framework
        self.config_logger = config_logger
        self.config_persistence = config_persistence
        self.config_providers = config_providers
        self.config_events = config_events
        self.options = options

    @staticmethod
    @meiga
    def from_filename(filename: str) -> Result[Any, Error]:
        if not os.path.isfile(filename):
            return Failure(ConfigFileNotFoundError(filename))
        try:
            petisco_yml_folder = os.path.dirname(filename)
            with open(filename) as file:
                petisco_dict = yaml.load(file, Loader=yaml.FullLoader)
                petisco_dict["petisco_yml_folder"] = petisco_yml_folder
                config = Config.from_dict(petisco_dict).unwrap_or_return()
                return Success(config)
        except (ParserError, ScannerError) as e:
            message = f"Error loading {filename} file: {repr(e.__class__)} {e} | {traceback.format_exc()}"
            return Failure(ConfigFileNotValidError(message))

    @staticmethod
    @meiga
    def from_dict(yaml_dict: Dict) -> Result[Any, Error]:
        petisco_yml_folder = yaml_dict.get("petisco_yml_folder")
        app_config = yaml_dict.get("app")
        app_name = app_config.get("name")
        app_version = Config.get_version(
            petisco_yml_folder, app_config.get("version")
        ).unwrap_or_return()

        config_tasks = ConfigTasks.from_dict(yaml_dict.get("tasks"))

        config_framework = ConfigFramework.from_dict(yaml_dict.get("framework"))

        config_logger = Config.get_config_logger(
            yaml_dict.get("logger")
        ).unwrap_or_return()

        config_persistence = ConfigPersistence.from_dict(yaml_dict.get("persistence"))

        config_providers = ConfigProviders.from_dict(yaml_dict.get("providers"))

        config_events = ConfigEvents.from_dict(yaml_dict.get("events"))

        options = app_config.get("options")

        return Success(
            Config(
                petisco_yml_folder=petisco_yml_folder,
                app_name=app_name,
                app_version=app_version,
                config_tasks=config_tasks,
                config_framework=config_framework,
                config_logger=config_logger,
                config_providers=config_providers,
                config_events=config_events,
                config_persistence=config_persistence,
                options=options,
            )
        )

    @staticmethod
    def get_version(petisco_yml_folder, config_version) -> Result[str, Error]:
        if isinstance(config_version, str):
            return Success(config_version)
        else:
            version_filename = config_version.get("from_file")
            if not version_filename:
                Failure(
                    ConfigFileNotValidError(
                        "If you don't specify the version directly, you must associate a file with from_file key"
                    )
                )

            version_filename = f"{petisco_yml_folder}/{version_filename}"
            if not os.path.isfile(version_filename):
                return Failure(
                    ConfigFileNotValidError(
                        f"Version from_file does not exist ({version_filename})"
                    )
                )
            version = open(version_filename, "r").read()
            return Success(version.rstrip("\n"))

    @staticmethod
    def get_config_logger(config_logger_dict: Dict) -> Result[ConfigLogger, Error]:

        if not config_logger_dict:
            return Success(ConfigLogger())
        else:
            return Success(ConfigLogger.from_dict(config_logger_dict))

    def get_logger(self) -> ILogger:
        if not self.config_logger or self.config_logger.selected_logger != "logging":
            return NotImplementedLogger()
        else:
            return LoggingBasedLogger(
                logger_name=self.config_logger.name,
                format=self.config_logger.format,
                config=self.config_logger.config,
            )

    def get_application(self) -> IApplication:
        if not self.config_framework:
            return None

        if self.config_framework.selected_framework != "flask":
            raise TypeError(
                f"Selected framework ({self.config_framework.selected_framework}) is not allowed. Try with flask."
            )
        else:
            if not self.config_framework.port_env:
                port = self.config_framework.port
            else:
                port = (
                    self.config_framework.port
                    if not os.environ.get(self.config_framework.port_env)
                    else os.environ.get(self.config_framework.port_env)
                )

            from petisco.frameworks.flask.application.flask_application import (
                FlaskApplication,
            )

            conf_file_split: List = self.config_framework.config_file.split("/")

            if len(conf_file_split) == 1:
                swagger_dir = self.petisco_yml_folder
                config_file = conf_file_split[0]
            else:
                swagger_dir, config_file = conf_file_split

            return FlaskApplication(
                application_name=self.app_name,
                swagger_dir=swagger_dir,
                config_file=config_file,
                port=port,
            )



import os

import pytest

from petisco.application.petisco import Petisco


@pytest.fixture
def petisco_sql_database():
    if not Petisco.get_instance().persistence_configured:
        yield
    else:
        from sqlalchemy import create_engine
        from petisco.persistence.sqlalchemy.sqlalchemy_persistence import (
            SqlAlchemyPersistence,
        )

        sql_database = os.environ.get("SQL_DATABASE")
        if not sql_database:
            raise RuntimeError(
                "Please add required SQL_DATABASE env variable (e.g pytest.ini, pytest.end2end.ini)"
            )

        Base = SqlAlchemyPersistence.get_instance().base
        Session = SqlAlchemyPersistence.get_instance().session
        connection = f"sqlite:///{sql_database}"
        engine = create_engine(connection)
        Base.metadata.create_all(engine)

        yield

        session = Session()
        session.rollback()
        session.close()
        Base.metadata.drop_all(bind=engine)
        os.remove(sql_database)



from typing import Dict


def raise_petisco_config_exception(kdict: Dict, field_name=None):
    if field_name:
        field_name = f" ({field_name})"
    else:
        field_name = ""
    raise TypeError(
        f"Petisco Config Error: Given input{field_name} is not valid. Please check it {kdict}"
    )
